@@ -4,6 +4,7 @@ qa_utils: Tools for use in quality assurance testing.
 
 import contextlib
 import datetime
+import hashlib
 import io
 import os
 import pytest
@@ -12,9 +13,10 @@ import re
 import time
 import toml
 import uuid
+import warnings
 
 from json import dumps as json_dumps, loads as json_loads
-from .misc_utils import PRINT, ignored, Retry, CustomizableProperty, getattr_customized
+from .misc_utils import PRINT, ignored, Retry, CustomizableProperty, getattr_customized, remove_prefix
 
 
 def show_elapsed_time(start, end):
@@ -401,17 +403,17 @@ class MockFileSystem:
         if not self.files.pop(file, None):
             raise FileNotFoundError("No such file or directory: %s" % file)
 
-    def open(self, file, mode='r'):
+    def open(self, file, mode='r', encoding=None):
         if FILE_SYSTEM_VERBOSE:
             print("Opening %r in mode %r." % (file, mode))
         if mode == 'w':
-            return self._open_for_write(file_system=self, file=file, binary=False)
+            return self._open_for_write(file_system=self, file=file, binary=False, encoding=encoding)
         elif mode == 'wb':
-            return self._open_for_write(file_system=self, file=file, binary=True)
+            return self._open_for_write(file_system=self, file=file, binary=True, encoding=encoding)
         elif mode == 'r':
-            return self._open_for_read(file, binary=False)
+            return self._open_for_read(file, binary=False, encoding=encoding)
         elif mode == 'rb':
-            return self._open_for_read(file, binary=True)
+            return self._open_for_read(file, binary=True, encoding=encoding)
         else:
             raise AssertionError("Mocked io.open doesn't handle mode=%r." % mode)
 
@@ -600,12 +602,16 @@ class MockBoto3:
 
     def __init__(self, **override_mappings):
         self._mappings = dict(self._default_mappings(), **override_mappings)
+        self.shared_reality = {}
+
+    def resource(self, key, **kwargs):
+        return self._mappings[key](boto3=self, **kwargs)
 
     def client(self, kind, **kwargs):
         mapped_class = self._mappings.get(kind)
         if not mapped_class:
             raise NotImplementedError("Unsupported boto3 mock kind:", kind)
-        return mapped_class(**kwargs)
+        return mapped_class(boto3=self, **kwargs)
 
 
 class MockBotoS3Client:
@@ -616,19 +622,30 @@ class MockBotoS3Client:
     MOCK_STATIC_FILES = {}
     MOCK_REQUIRED_ARGUMENTS = {}
 
-    def __init__(self, region_name=None, mock_other_required_arguments=None, mock_s3_files=None):
+    DEFAULT_STORAGE_CLASS = "STANDARD"
+
+    def __init__(self, region_name=None, mock_other_required_arguments=None, mock_s3_files=None,
+                 storage_class=None, boto3=None):
+        self.boto3 = boto3 or MockBoto3()
         if region_name not in (None, 'us-east-1'):
             raise ValueError("Unexpected region:", region_name)
 
-        files = self.MOCK_STATIC_FILES.copy()
-        for name, content in mock_s3_files or {}:
-            files[name] = content
-        self.s3_files = MockFileSystem(files=files)
+        files_cache_marker = '_s3_file_data'
+        shared_reality = self.boto3.shared_reality
+        s3_files = shared_reality.get(files_cache_marker)
+        if s3_files is None:
+            files = self.MOCK_STATIC_FILES.copy()
+            for name, content in mock_s3_files or {}:
+                files[name] = content
+            shared_reality[files_cache_marker] = s3_files = MockFileSystem(files=files)
+        self.s3_files = s3_files
 
         other_required_arguments = self.MOCK_REQUIRED_ARGUMENTS.copy()
         for name, content in mock_other_required_arguments or {}:
             other_required_arguments[name] = content
         self.other_required_arguments = other_required_arguments
+        self.storage_class = storage_class or self.DEFAULT_STORAGE_CLASS
+
 
     def upload_fileobj(self, Fileobj, Bucket, Key, **kwargs):  # noqa - Uppercase argument names are chosen by AWS
         if kwargs != self.other_required_arguments:
@@ -664,24 +681,172 @@ class MockBotoS3Client:
         with io.open(Filename, 'wb') as fp:
             self.download_fileobj(Bucket=Bucket, Key=Key, Fileobj=fp)
 
-    def get_object(self, Bucket, Key, **kwargs):
+    def get_object(self, Bucket, Key, **kwargs):  # noqa - Uppercase argument names are chosen by AWS
         if kwargs != self.other_required_arguments:
             raise MockKeysNotImplemented("get_object", kwargs.keys())
 
+        head_metadata = self.head_object(Bucket=Bucket, Key=Key, **kwargs)
+
+        pseudo_filename = os.path.join(Bucket, Key)
+
+        return dict(head_metadata,
+                    Body=self.s3_files.open(pseudo_filename, 'rb'))
+
+    PUT_OBJECT_CONTENT_TYPES = {
+        "text/html": [".html"],
+        "image/png": [".png"],
+        "application/json": [".json"],
+        "text/plain": [".txt", ".text"],
+        "binary/octet-stream": [".fo"],
+    }
+
+    def put_object(self, *, Bucket, Key, Body, ContentType=None, **kwargs):  # noqa - Uppercase argument names are chosen by AWS
+        if ContentType is not None:
+            exts = self.PUT_OBJECT_CONTENT_TYPES.get(ContentType)
+            assert exts, "Unimplemented mock .put_object content type %s for Key=%s" % (ContentType, Key)
+            assert any(Key.endswith(ext) for ext in exts), (
+                    "mock .put_object expects Key=%s to end in one of %s for ContentType=%s" % (Key, exts, ContentType))
+        assert not kwargs, "put_object mock doesn't support %s." % kwargs
+        self.s3_files.files[Bucket + "/" + Key] = Body
         return {
-            "Body": self.s3_files.open(os.path.join(Bucket, Key), 'rb'),
+            'ETag': self._content_etag(Body)
         }
 
+    @staticmethod
+    def _content_etag(content):
+        return hashlib.md5(content).hexdigest()
+
+    def Bucket(self, name):  # noQA - AWS function naming style
+        return MockBotoS3Bucket(s3=self, name=name)
+
+    def head_object(self, Bucket, Key, **kwargs):  # noQA - AWS argument naming style
+        if kwargs != self.other_required_arguments:
+            raise MockKeysNotImplemented("get_object", kwargs.keys())
+
+        pseudo_filename = os.path.join(Bucket, Key)
+
+        if self.s3_files.exists(pseudo_filename):
+            content = self.s3_files.files[pseudo_filename]
+            return {
+                'Bucket': Bucket,
+                'Key': Key,
+                'ETag': self._content_etag(content),
+                'ContentLength': len(content),
+                # Numerous others, but this is enough to make the dictionary non-empty and to satisfy some of our tools
+            }
+        else:
+            # I would need to research what specific error is needed here and hwen,
+            # since it might be a 404 (not found) or a 403 (permissions), depending on various details.
+            # For now, just fail in any way since maybe our code doesn't care.
+            raise Exception("Mock File Not Found")
+
+    def list_objects(self, Bucket, Prefix=None):  # noQA - AWS argument naming style
+        # Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects
+        bucket_prefix = Bucket + "/"
+        bucket_prefix_length = len(bucket_prefix)
+        search_prefix = bucket_prefix + (Prefix or '')
+        found = []
+        for filename, content in self.s3_files.files.items():
+            if filename.startswith(search_prefix):
+                found.append({
+                    'Key': filename[bucket_prefix_length:],
+                    'ETag': self._content_etag(content),
+                    # "LastModified": ...,
+                    # "Owner": {"DisplayName": ..., "ID"...},
+                    "Size": len(content),
+                    "StorageClass": self.storage_class,
+                })
+        return {
+            # "CommonPrefixes": {"Prefix": ...},
+            "Contents": found,
+            # "ContinuationToken": ...,
+            # "Delimiter": ...,
+            # "EncodingType": ...,
+            # "KeyCount": ...,
+            "IsTruncated": False,
+            # "MaxKeys": ...,
+            # "NextContinuationToken": ...,
+            "Name": Bucket,
+            "Prefix": Prefix,
+            # "StartAfter": ...,
+        }
+
+
+class MockBotoS3Bucket:
+
+    def __init__(self, name, s3=None):
+        assert s3, "missing s3"
+        self.s3 = s3
+        self.name = name
+        self.objects = MockBotoS3BucketObjects(bucket=self)
+
+    def _delete(self, delete_bucket_too=False):
+        prefix = self.name + "/"
+        files = self.s3.s3_files.files
+        to_delete = set()
+        for pseudo_filename, _ in [files.items()]:
+            if pseudo_filename.startswith(prefix):
+                if pseudo_filename != prefix:
+                    to_delete.add(pseudo_filename)
+        for pseudo_filename in to_delete:
+            del files[pseudo_filename]
+        if not delete_bucket_too:
+            files[prefix] = b''
+        # TODO: Does anything need to be returned here?
+
+    def _keys(self):
+        found = False
+        keys = set()  # In real S3, this would be cached info, but for testing we just create it on demand
+        prefix = self.name + "/"
+        for pseudo_filename, content in self.s3.s3_files.files.items():
+            if pseudo_filename.startswith(prefix):
+                found = True
+                key = remove_prefix(prefix, pseudo_filename)
+                if key != prefix:
+                    keys.add(key)
+        if not keys:
+            if not found:
+                # It's OK if we only found "<bucketname>/", which is an empty, but not missing bucket
+                raise Exception("NoSuchBucket")
+        return sorted(keys)
+
+    def delete(self):
+        self._delete(delete_bucket_too=True)
+
+    def _all(self):
+        """A callback for <bucket>.objects.all()"""
+        return [self.s3.head_object(Bucket=self.name, Key=key)
+                for key in self._keys()]
+
+
+class MockBotoS3ObjectSummary:
+    # Not sure if we need to expose the strucutre of this yet. -kmp 13-Jan-2021
+    def __init__(self, **attributes):
+        self._attributes = attributes
+
+
+class MockBotoS3BucketObjects:
+
+    def __init__(self, bucket):
+        self.bucket = bucket
+
+    def all(self):
+        return self.bucket._all()  # noQA - we are effectively a friend of this instance and are intended to call this.
+
+    def delete(self):
+        return self.bucket._delete(delete_bucket_too=False)  # noQA - we are effectively a friend of this instance
+
 
 class MockBotoSQSClient:
     """
     This is a mock of certain SQS functionality.
     """
 
-    def __init__(self, region_name=None):
+    def __init__(self, region_name=None, boto3=None):
         if region_name not in (None, 'us-east-1'):
             raise RuntimeError("Unexpected region:", region_name)
         self._mock_queue_name_seen = None
+        self.boto3 = boto3 or MockBoto3()
 
     def check_mock_queue_url_consistency(self, queue_url):
         __tracebackhide__ = True
@@ -775,6 +940,12 @@ class VersionChecker:
     PYPROJECT = CustomizableProperty('PYPROJECT', description="The repository-relative name of the pyproject file.")
     CHANGELOG = CustomizableProperty('CHANGELOG', description="The repository-relative name of the change log.")
 
+    # I wanted to use pytest.PytestConfigWarning, but that creates a dependency
+    # on particular versions of pytest, and we don't export a delivery
+    # constraint of a particular pytest version. So RuntimeWarning is a
+    # safer setting for now. -kmp 14-Jan-2021
+    WARNING_CATEGORY = RuntimeWarning
+
     @classmethod
     def check_version(cls):
         version = cls._check_version()
@@ -823,10 +994,12 @@ class VersionChecker:
                     versions.append(m.group(1))
 
         assert versions, "No version info was parsed from %s" % changelog_file
+
         # Might be sorted top to bottom or bottom to top, but ultimately the current version should be first or last.
-        assert versions[0] == version or versions[-1] == version, (
-                "Missing entry for version %s in %s." % (version, changelog_file)
-        )
+        if versions[0] != version and versions[-1] != version:
+            warnings.warn("Missing entry for version %s in %s." % (version, changelog_file),
+                          category=cls.WARNING_CATEGORY, stacklevel=2)
+            return
 
 
 def raises_regexp(error_class, pattern):
